@@ -19,6 +19,7 @@ along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
 
 from libopensesame.py3compat import *
 import uuid
+import os
 import ast
 import time
 import json
@@ -26,7 +27,14 @@ import inspect
 import pickle
 from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
 from qtpy.QtWidgets import QApplication
+from libopensesame.oslogging import oslogger
 from qtconsole.rich_jupyter_widget import RichJupyterWidget
+try:
+    from base64 import decodebytes
+except ImportError:  # Python 2
+    from base64 import decodestring as decodebytes
+from libqtopensesame.misc.translate import translation_context
+_ = translation_context(u'JupyterConsole', category=u'extension')
 
 
 GLOBAL_EXPR = u'''{
@@ -59,6 +67,13 @@ not inspect.ismodule(val)
 '''
 
 
+class FailedToGetWorkspaceVariable(Exception):
+    
+    def __repr__(self):
+        
+        return _('Failed to get workspace variable')
+
+
 class TransparentJupyterWidget(RichJupyterWidget, BaseSubcomponent):
 
     def __init__(self, jupyter_console):
@@ -69,19 +84,71 @@ class TransparentJupyterWidget(RichJupyterWidget, BaseSubcomponent):
         self.setup(jupyter_console)
         self.executed.connect(self._on_executed)
         self.executing.connect(self._on_executing)
+        self._last_executed = None
+        # It is possible for _on_executing() to be called again, before
+        # _on_executed() is called. We therefore keep track of how many times
+        # _on_executing() has been called, and only release the busy status
+        # when this counter is set back to 0.
+        self._executing_counter = 0
+        
+    def execute(self, *args, **kwargs):
+        
+        self._last_executed = args[0] if args else kwargs.get('source', None)
+        super(TransparentJupyterWidget, self).execute(*args, **kwargs)
 
     def _on_executing(self):
 
+        self._executing_counter += 1
         self._jupyter_console.set_busy(True)
 
     def _on_executed(self):
 
+        if self._executing_counter:
+            self._executing_counter -= 1
+        if self._executing_counter:
+            return
         self._jupyter_console.set_busy(False)
         self.extension_manager.fire(
             u'workspace_update',
             name=self._name,
             workspace_func=self.get_workspace_globals
         )
+        
+    def running(self):
+        
+        return self._executing_counter > 0
+    
+    def reset(self, clear=False):
+        
+        super(TransparentJupyterWidget, self).reset(clear=clear)
+        self._executing_counter = 0
+        self._jupyter_console.set_busy(False)
+        
+    def _handle_display_data(self, msg):
+        """Reimplemented to handle communications between the figure explorer
+        and the kernel. Inspired by Spyder's `figurebrowser.py`.
+        """
+        img = None
+        data = msg['content']['data']
+        if 'image/svg+xml' in data:
+            fmt = 'image/svg+xml'
+            img = data['image/svg+xml']
+        elif 'image/png' in data:
+            # PNG data is base64 encoded as it passes over the network
+            # in a JSON structure so we decode it.
+            fmt = 'image/png'
+            img = decodebytes(data['image/png'].encode('ascii'))
+        elif 'image/jpeg' in data and self._jpg_supported:
+            fmt = 'image/jpeg'
+            img = decodebytes(data['image/jpeg'].encode('ascii'))
+        if img is not None:
+            self.extension_manager.fire(
+                'image_annotations_new',
+                img=img,
+                fmt=fmt,
+                code=self._last_executed
+            )
+        return super(TransparentJupyterWidget, self)._handle_display_data(msg)
 
 
 class OutprocessJupyterWidget(TransparentJupyterWidget):
@@ -107,7 +174,7 @@ class OutprocessJupyterWidget(TransparentJupyterWidget):
         
         key = str(uuid.uuid4())
         self._kernel_client.execute(
-            u'import inspect',
+            u'import os, inspect',
             silent=True,
             user_expressions={
                 key: expr
@@ -132,6 +199,17 @@ class OutprocessJupyterWidget(TransparentJupyterWidget):
     def get_workspace_globals(self):
 
         return self._silent_execute(GLOBAL_EXPR)
+        
+    @property
+    def pid(self):
+        
+        for i in range(10):
+            pid = self._silent_execute('os.getpid()')
+            if isinstance(pid, int):
+                return pid
+            time.sleep(0.1)
+        oslogger.warning('failed to get pid')
+        return -1
 
     def list_workspace_globals(self):
 
@@ -172,17 +250,22 @@ class OutprocessJupyterWidget(TransparentJupyterWidget):
             time.sleep(0.05)
             QApplication.processEvents()
         else:
-            return None
+            return FailedToGetWorkspaceVariable()
         reply = self._user_expressions[key].get(
             u'data',
             {}
         ).get(u'text/plain', None)
         if reply is None:
-            return None
+            return FailedToGetWorkspaceVariable()
         try:
             return pickle.loads(eval(reply))
-        except (TypeError, pickle.UnpicklingError, RecursionError):
-            return None
+        except (
+            TypeError,
+            pickle.UnpicklingError,
+            RecursionError,
+            ModuleNotFoundError
+        ):
+            return FailedToGetWorkspaceVariable()
 
 
 class InprocessJupyterWidget(TransparentJupyterWidget):
@@ -212,6 +295,11 @@ class InprocessJupyterWidget(TransparentJupyterWidget):
             not inspect.isclass(val) and
             not inspect.ismodule(val)
         }
+    
+    @property
+    def pid(self):
+
+        return os.getpid()
         
     def list_workspace_globals(self):
         
